
/*
**  jazz - a midi sequencer for Linux
**
**  Copyright (C) 1994 Andreas Voss (andreas@avix.rni.sub.org)
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

// 
// $Id: player.cc,v 1.6 1994/06/25 12:24:08 andreas Exp andreas $
// 

#include "player.h"
#include "trackwin.h"


tPlayer *Midi = 0;

tPlayer::tPlayer(tSong *song)
{
  Song = song;
  OutClock = 0;
  DeltaClock = 480;
  Playing = 0;
}

tPlayer::~tPlayer()
{
}

void tPlayer::StartPlay(long Clock)
{
  int i;
  OutClock = Clock;
  DeltaClock = 480;
  PlayBuffer.Clear();
  RecdBuffer.Clear();

  // Send Volume, Pan, Chorus, etc
  for (i = 0; i < Song->nTracks; i++)
  {
    tTrack *t = Song->GetTrack(i);
    if (t->Patch)
      OutNow(t->Patch);
    if (t->Volume)
      OutNow(t->Volume);
    if (t->Pan)
      OutNow(t->Pan);
    if (t->Reverb)
      OutNow(t->Reverb);
    if (t->Chorus)
      OutNow(t->Chorus);
    if (t->Speed)
      OutNow(t->Speed);
  }
  Notify();
  Start(50);	// wxTimer: call Notify every 100 ms
  Playing = 1;
}

void tPlayer::StopPlay()
{
  Stop();	// stop wxTimer
  Playing = 0;
  RecdBuffer.Keyoff2Length();
  TrackWin->NewPlayPosition(-1L);
}

void tPlayer::Notify()
{
  long Now = GetRealTimeClock();
  // buffer-events eaten up by device
  if (PlayBuffer.GetFirstClock() >= OutClock)
    // time to put more events
    if (Now + DeltaClock > OutClock)
    {
      Song->MergeTracks(OutClock, Now + DeltaClock, &PlayBuffer);
      OutClock = Now + DeltaClock;
      PlayBuffer.Length2Keyoff();
      TrackWin->NewPlayPosition(Now);
    }
  FlushToDevice();
}


void tPlayer::FlushToDevice()
// try to send all events up to OutClock to device
{
  int BufferFull = 0;

  tEventIterator Iterator(&PlayBuffer);
  tEvent *e = Iterator.Range(0, OutClock);
  while (!BufferFull && e) 
  {
    if (OutEvent(e) != 0)
      BufferFull = 1;
    else
    {
      e->Kill();
      e = Iterator.Next();
    }
  }
  if (!BufferFull)
    OutBreak();
  PlayBuffer.Cleanup(0);
}

void tPlayer::AllNotesOff(int Reset)
{
  tControl CtrlRes(0, 0, 0x79, 0);
  tControl NoteOff(0, 0, 0x7B, 0);
  tControl Volume (0, 0, 0x07, 100);
  tControl Pan    (0, 0, 0x0A, 64);
  tControl Reverb (0, 0, 0x5B, 0);
  tControl Chorus (0, 0, 0x5D, 0);
  tProgram Program(0, 0, 0);
  tPitch   Pitch  (0, 0, 0);
  for (int i = 0; i < 16; i++)
  {
    NoteOff.Channel = i; OutNow(&NoteOff);
    Pitch.Channel = i;   OutNow(&Pitch);
    CtrlRes.Channel = i; OutNow(&CtrlRes);
    if (Reset)
    {
      Volume.Channel = i;  OutNow(&Volume);
      Pan.Channel = i;     OutNow(&Pan);
      Reverb.Channel = i;  OutNow(&Reverb);
      Chorus.Channel = i;  OutNow(&Chorus);
      Program.Channel = i; OutNow(&Program);
    }
  }
}




// ----------------------------------------------------------------------------------------------
// MPU-Player
// ----------------------------------------------------------------------------------------------

tMpuPlayer::tMpuPlayer(tSong *song)
  : tPlayer(song)
{
  dev = open(MPUDEVICE, O_RDWR | O_NONBLOCK);
}



tMpuPlayer::~tMpuPlayer()
{
  close(dev);
}


int tMpuPlayer::Installed()
{
  return dev >= 0;
}


#if DB_WRITE
int dwrite(int dev, const char *buf, int size)
{
  int i, written;
  //dev = 2;	// stderr
  written = write(dev, buf, size);
  if (written < 0)
  {
    printf("writing failed\n");
    return written;
  }
  printf("W: ");
  for (i = 0; i < written; i++)
    printf("%02x ", (unsigned char)buf[i]);
  putchar('\n');
  if (written != size)
  {
    printf("L: ");
    for (i = written; i < size; i++)
      printf("%02x ", (unsigned char)buf[i]);
    putchar('\n');
  }
  fflush(stdout);
  return written;
}
#endif

void tMpuPlayer::StartPlay(long Clock)
{
  static const char play[] = 
  {
    CMD+1, 0x34,			    		/* timing byte always */
    CMD+1, 0x8e,					/* conductor off */
    CMD+1, 0x8c,					/* don't send measures while recording */
    /* CMD+1, 0x95, */ 					/* send clock to host instead */
    CMD+1, 0x87,  					/* pitch+controller enabled */
    CMD+1, 0x98, CMD+1, 0x9a, CMD+1, 0x9c, CMD+1, 0x9e,	/* channel-ref-tables off */
    CMD+1, 0xec, DAT+1, 0x01, CMD+1, 0xb8, CMD+1, 0x2a  /* start record/play */
  };

  PlyBytes.Clear();
  PlyBytes.Clock = Clock;

  RecBytes.Clear();
  RecBytes.Clock = Clock;

  // Setup Timebase
  char timebase[2];
  timebase[0] = CMD+1;
  switch (Song->TicksPerQuarter)
  {
    case  48: timebase[1] = 0xc2; break;
    case  72: timebase[1] = 0xc3; break;
    case  96: timebase[1] = 0xc4; break;
    case 120: timebase[1] = 0xc5; break;
    case 144: timebase[1] = 0xc6; break;
    case 168: timebase[1] = 0xc7; break;
    case 192: timebase[1] = 0xc8; break;
    default : timebase[1] = 0xc5; break;
  }
  dwrite(dev, timebase, 2);

  tPlayer::StartPlay(Clock);
  dwrite(dev, play, sizeof(play));
}



void tMpuPlayer::StopPlay()
{
  static const char stop = RES;
  tPlayer::StopPlay();
  dwrite(dev, &stop, 1);
  PlyBytes.Clear();
  AllNotesOff();
}




int tMpuPlayer::OutEvent(tEvent *e)
{
  if (!PlyBytes.WriteFile(dev))
    return 1;	// buffer full

  int Stat = e->Stat;

  switch (Stat)
  {
    case StatKeyOff:
    case StatKeyOn:
    case StatAftertouch:
    case StatControl:
    case StatProgram:
    case StatMono:
    case StatPitch:
      {
	tGetMidiBytes midi;
	int i;

	e->Write(midi);
	Stat = midi.Buffer[0]; // Status + Channel

	OutBreak(e->Clock - PlyBytes.Clock);
	long Time = e->Clock - PlyBytes.Clock;
	assert(Time < 240);

	if (Stat != PlyBytes.RunningStatus)
	{
	  PlyBytes.Put(TRK + midi.nBytes + 1);
	  PlyBytes.Put(Time);
	  PlyBytes.Put(Stat);
	  PlyBytes.RunningStatus = Stat;
	}
	else
	{
	  PlyBytes.Put(TRK + midi.nBytes);
	  PlyBytes.Put(Time);
	}
	for (i = 1; i < midi.nBytes; i++)
	  PlyBytes.Put(midi.Buffer[i]);

	PlyBytes.Clock = e->Clock;
	return 0;
      }

    default:	// Meterchange etc
      return 0;
      break;
  }
}


void tMpuPlayer::OutBreak()
{
  // wait < OutClock
  if (!PlyBytes.WriteFile(dev))
    return;

  long Delta = OutClock - PlyBytes.Clock;
  (void)OutBreak(Delta);
  PlyBytes.WriteFile(dev);
}


long tMpuPlayer::OutBreak(long Delta)
{
  while (Delta >= 240)
  {
    PlyBytes.Put(TRK+1);
    PlyBytes.Put(0xf8);
    Delta -= 240;
    PlyBytes.Clock += 240;
  }
  return Delta;
}



void tMpuPlayer::OutNow(tEvent *e)
{
  char buf[20];
  int i, n = 0;
  tGetMidiBytes midi;
  if (e->Write(midi) == 0)
  {
    buf[n++] = CMD+1;
    buf[n++] = 0xd1;
    buf[n++] = DAT+midi.nBytes;
    for (i = 0; i < midi.nBytes; i++)
      buf[n++] = midi.Buffer[i];
    dwrite(dev, buf, n);
  }

  else	// special event
  {
    switch (e->Stat)
    {
      case StatSetTempo:
	{
	  char cmd[20];
	  tSetTempo *s = (tSetTempo *)e;
	  int bpm = s->GetBPM();
	  cmd[0] = CMD+1;
	  cmd[1] = 0xE0;
	  cmd[2] = DAT+1;
	  cmd[3] = (char)bpm;
	  dwrite(dev, cmd, 4);
	}
	break;

      default:	// ignore others
        break;
    }
  }
}



long tMpuPlayer::GetRealTimeClock()
{
  int c;
  RecBytes.ReadFile(dev);
  while ((c = RecBytes.Get()) >= 0)
  {
    if (c == 0xf8)
      RecBytes.Clock += 240;
    else if (c < 0xf0)
    {
      // timing byte
      RecBytes.Clock += c;
      c = RecBytes.Get();
      if (c < 0xf0)  // Voice-Message
      {
        unsigned char c1, c2;
        int Channel;
        tEvent *e = 0;

        if (c & 0x80)
        {
          RecBytes.RunningStatus = c;
          c1 = RecBytes.Get();
	}
	else
	  c1 = c;

	Channel = RecBytes.RunningStatus & 0x0f;
	switch (RecBytes.RunningStatus & 0xf0)
	{
	  case StatKeyOff:
	    (void)RecBytes.Get();
	    e = new tKeyOff(RecBytes.Clock, Channel, c1);
	    break;

	  case StatKeyOn:
	    c2 = RecBytes.Get();
	    if (!c2)
	      e = new tKeyOff(RecBytes.Clock, Channel, c1);
	    else
	      e = new tKeyOn(RecBytes.Clock, Channel, c1, c2);
	    break;
#if 0
	  case StatAftertouch:
	    c2 = RecBytes.Get();
	    e  = new tAftertouch(RecBytes.Clock, Channel, c1, c2);
	    break;

	  case StatMono:
	    e  = new tMono(RecBytes.Clock, Channel, c1);
	    break;
#endif

	  case StatControl:
	    c2 = RecBytes.Get();
	    e  = new tControl(RecBytes.Clock, Channel, c1, c2);
	    break;

	  case StatProgram:
	    e  = new tProgram(RecBytes.Clock, Channel, c1);
	    break;
	    break;

	  case StatPitch:
	    c2 = RecBytes.Get();
	    e  = new tPitch(RecBytes.Clock, Channel, c1, c2);
	    break;

	  default:
	    printf("unrecognized MIDI Status %02x, %02x", (unsigned char)RecBytes.RunningStatus, c1);
	    break;

	}
	if (e)
	  RecdBuffer.Put(e);
      }
      else
        printf("unrecognized Status after Timing Byte: %02x\n", c);
    }
    else
      printf("unrecognized Status (no timing byte): %02x\n", c);
  }
  return RecBytes.Clock;
}


