
/*
**  jazz - a midi sequencer for Linux
**
**  Copyright (C) 1994 Andreas Voss (andreas@avix.rni.sub.org)
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

// 
// $Id: track.cc,v 1.7 1994/06/25 12:24:08 andreas Exp andreas $
// 

#include "track.h"
#include "util.h"
#include "player.h"
#include "jazz.h"

#include <stdlib.h>

#include <assert.h>


tEventArray::tEventArray()
{
  nEvents = 0;

  MaxEvents = 0;
  Events = 0;

  Clear();
}


void tEventArray::Clear()
{
  for (int i = 0; i < nEvents; i++)
    delete Events[i];
  nEvents = 0;

  Name   = 0;
  Patch  = 0;
  Volume = 0;
  Pan    = 0;
  Reverb = 0;
  Chorus = 0;
  Speed = 0;
  Channel = 1;

  if (Events)
    free(Events);
  Events = 0;
  MaxEvents = 0;
}


tEventArray::~tEventArray()
{
  Clear();
}

void tEventArray::Resize()
{
  MaxEvents += 20;
  Events = (tEvent **)realloc(Events, MaxEvents * sizeof(tEvent *));
  assert(Events);	// crash!!
}


void tEventArray::Put(tEvent *e)
{
  if (nEvents >= MaxEvents)
    Resize();
  Events[nEvents++] = e;
}


static int compare(void *p1, void *p2)
{
  tEvent *e1 = *(tEvent **)p1;
  tEvent *e2 = *(tEvent **)p2;
  return e1->Compare(*e2);
}


void tEventArray::Sort()
{
  qsort(Events, nEvents, sizeof(tEvent *), compare);
}



void tEventArray::Cleanup(tUndoBuffer *UndoBuffer)
{
  tEvent *e;
  tControl *c;

  Sort();	// Alle Killed-Events sind am Ende des Arrays

  Name = 0;
  Patch = 0;
  Speed = 0;
  Volume = 0;
  Pan = 0;
  Reverb = 0;
  Chorus = 0;

  for (int i = 0; i < nEvents; i++)
  {
    if ((e = Events[i])->IsKilled())
    {
      if (UndoBuffer)
      {
	for (int j = i; j < nEvents; j++)
	  UndoBuffer->Put(Events[j]);
      }
      else
      {
	for (int j = i; j < nEvents; j++)
	  delete Events[j];
      }
      nEvents = i;
      break;
    }

    if (!Name) 
      Name = e->IsTrackName();
    if (!Patch)
      Patch = e->IsProgram();
    if (!Speed)
      Speed = e->IsSetTempo();
    if ((c = e->IsControl()) != 0)
    {
      switch (c->Control)
      {
        case 0x07: if (!Volume) Volume = c; break;
        case 0x0a: if (!Pan)    Pan    = c; break;
        case 0x5b: if (!Reverb) Reverb = c; break;
        case 0x5d: if (!Chorus) Chorus = c; break;
      }
    }
  }
}


void tEventArray::Length2Keyoff()
{
  int n = nEvents;
  for (int i = 0; i < n; i++)
  {
    tKeyOn *on;
    if ((on = Events[i]->IsKeyOn()) != 0 && on->Length != 0)
    {
      tEvent *of = new tKeyOff(on->Clock + on->Length, on->Channel, on->Key);
      on->Length = 0;
      Put(of);
    }
  }
  Sort();
}



void tEventArray::Keyoff2Length()
{
  for (int i = 1; i < nEvents; i++)
  {
    tKeyOff *of;
    if ((of = Events[i]->IsKeyOff()) != 0)
    {
      tEvent **e = &Events[i - 1];
      while (e >= Events)
      {
        tKeyOn *on = (*e)->IsKeyOn();
        if (on && on->Key == of->Key && on->Channel == of->Channel && on->Length == 0)
        {
          on->Length = of->Clock - on->Clock;
          if (on->Length <= 0L)
            on->Length = 1;
          of->Kill();
	  break;
	}
	-- e;
      }
    }
  }
  Cleanup(0);
}



void tEventArray::Write(tWriteBase &io)
{
  Length2Keyoff();
  io.NextTrack();
  for (int i = 0; i < nEvents; i++)
    Events[i]->Write(io);
  Keyoff2Length();
}



void tEventArray::Read(tReadBase &io)
{
  tEvent *e;
  Channel = 0;

  io.NextTrack();
  while ((e = io.Read()) != 0)
  {
    Put(e);
    if (!Channel && e->IsChannelEvent())
      Channel = e->IsChannelEvent()->Channel + 1;
  }
  if (!Channel)
    Channel = 1;

  Keyoff2Length();
}


long tEventArray::GetLastClock()
{
  if (!nEvents)
    return 0;
  return Events[nEvents-1]->Clock;
}

int tEventArray::IsEmpty() 
{ 
  return nEvents == 0; 
}

int tEventArray::GetFirstClock() 
{ 
  if (nEvents) 
    return Events[0]->Clock; 
  return LastClock; 
}

// ***********************************************************************
// Dialog
// ***********************************************************************

class tTrackDlg : public wxForm
{
  tTrack *trk;
  char *TrackName;
  tNamedChoice PatchChoice;
  long PatchNr;

 public:
  tTrackDlg::tTrackDlg(tTrack *t);
  void EditForm(wxPanel *panel);
  virtual void OnOk();
  virtual void OnCancel();
};


tTrackDlg::tTrackDlg(tTrack *t)
  : PatchChoice("Patch", VoiceNames, &PatchNr)
{
  trk = t;
}


void tTrackDlg::OnCancel()
{
  trk->DialogBox = 0;
  wxForm::OnCancel();
}


void tTrackDlg::OnOk()
{
  trk->DialogBox = 0;
  trk->SetName(TrackName);
  delete TrackName;
  PatchChoice.GetValue();
  trk->SetPatch(PatchNr);
  wxForm::OnOk();
}



void tTrackDlg::EditForm(wxPanel *panel)
{
  PatchNr   = trk->GetPatch();
  TrackName = copystring(trk->GetName());
  Add(wxMakeFormString("Trackname:", &TrackName, wxFORM_DEFAULT, 0,0,0, 100, -1));

  Add(wxMakeFormNewLine());
  Add(PatchChoice.mkFormItem(300, 300));
  Add(wxMakeFormNewLine());
  Add(wxMakeFormShort("Channel", &trk->Channel, wxFORM_DEFAULT,
                       new wxList(wxMakeConstraintRange(1.0, 16.0), 0)));

  AssociatePanel(panel);
}


void tTrack::Dialog(wxFrame *parent)
{
  tTrackDlg *dlg;
  if (DialogBox)
  {
    DialogBox->Show(TRUE);
    return;
  }
  DialogBox = new wxDialogBox(parent, "Track Settings", FALSE);
  dlg = new tTrackDlg(this);
  dlg->EditForm(DialogBox);
  DialogBox->Fit();
  DialogBox->Show(TRUE);
}

// ***********************************************************************
// tUndoBuffer
// ***********************************************************************

tUndoBuffer::tUndoBuffer()
{
}

void tUndoBuffer::Clear()
{
  Killed.Clear();
  Copies.nEvents = 0;
  Copies.Clear();
}

void tUndoBuffer::Put(tEvent *e)
{
  if (e->IsKilled())
    Killed.Put(e);
  else
    Copies.Put(e);
}

void tUndoBuffer::Debug()
{
  if (Killed.nEvents || Copies.nEvents)
  {
    int i;

    printf("tUndoBuffer.Killed:\n");
    for (i = 0; i < Killed.nEvents; i++)
    {
      tEvent *e = Killed.Events[i];
      printf("%p clk %08lx sta %02x\n", e, e->Clock, (unsigned char)e->Stat);
    }

    printf("tUndoBuffer.Copies:\n");
    for (i = 0; i < Copies.nEvents; i++)
    {
      tEvent *e = Copies.Events[i];
      printf("%p clk %08lx sta %02x\n", e, e->Clock, (unsigned char)e->Stat);
    }
    fflush(stdout);
  }
}


// ***********************************************************************
// tTrack
// ***********************************************************************


tTrack::tTrack()
  : tEventArray()
{
  iUndo = 0;
  DialogBox = 0;
  PatchNames = 0;
}


void tTrack::Merge(tEventArray *t)
{
  for (int i = 0; i < t->nEvents; i++)
    Put(t->Events[i]);
  t->nEvents = 0;
}


void tTrack::MergeRange(tEventArray *other, long FromClock, long ToClock, int Replace)
{
  // Erase destin
  if (Replace)
  {
    tEventIterator Erase(this);
    tEvent *e = Erase.Range(FromClock, ToClock);
    while (e)
    {
      Kill(e);
      e = Erase.Next();
    }
  }

  // Merge Recorded Events
  tEventIterator Copy(other);
  tEvent *e = Copy.Range(FromClock, ToClock);
  while (e)
  {

    #if 0
    {
      tGetMidiBytes mb;
      e->Write(mb);
      printf("ReplaceRange: Clock %ld, Stat %02x, Bytes %d: ", e->Clock, e->Stat, mb.nBytes);
      for (int i = 0; i < mb.nBytes; i++)
	printf("%02x ", mb.Buffer[i]);
      tKeyOn *on = e->IsKeyOn();
      if (on)
	printf("KeyOn, Length = %d", on->Length);
      tKeyOff *of = e->IsKeyOff();
      if (of)
	printf("KeyOFF********");

      putchar('\n');
      fflush(stdout);
    }
    #endif

    Put(e->Copy());
    e = Copy.Next();
  }
  Cleanup();
}


void tTrack::Cleanup()
{
  tEventArray::Cleanup(&UndoBuffers[iUndo]);
}



void tTrack::Undo()
{
  int i;
  tEventArray *Killed, *Copies;
  tUndoBuffer *u;

  u = &UndoBuffers[iUndo];
  Killed = &u->Killed;
  Copies = &u->Copies;

  for (i = 0; i < Killed->nEvents; i++)
  {
    tEvent *e = Killed->Events[i];
    e->UnKill();
    tEventArray::Put(e);
  }
  Killed->nEvents = 0;

  for (i = 0; i < Copies->nEvents; i++)
  {
    tEvent *e = Copies->Events[i];
    e->Kill();
  }
  Copies->nEvents = 0;

  iUndo = (iUndo - 1 + MaxUndo) % MaxUndo;
  tEventArray::Cleanup(0); 
}


void tTrack::NewUndoBuffer()
{
  iUndo = (iUndo + 1) % MaxUndo;
  UndoBuffers[iUndo].Clear();
};


void tTrack::Clear()
{
  tEventArray::Clear();
  for (int i = 0; i < MaxUndo; i++)
    UndoBuffers[i].Clear();
  State  = tsPlay;
}

// ----------------------- Name ------------------------------------

char *tTrack::GetName()
{
  if (Name)
    return (char *)Name->Data;
  return "";
}



void tTrack::SetName(char *str)
{
  if (Name)
    Kill(Name);
  if (strlen(str))
    Put(new tTrackName(0, (uchar *)str, strlen(str)));
  Cleanup();
}

// ------------------------  Volume ------------------------------

int tTrack::GetVolume()
{
  if (Volume)
    return Volume->Value + 1;
  return 0;
}

void tTrack::SetVolume(int Value)
{
  if (Volume)
    Kill(Volume);
  if (Value > 0)
  {
    tEvent *e = new tControl(0, Channel - 1, 0x07, Value - 1);
    Put(e);
    Midi->OutNow(e);
  }
  Cleanup();
}

// ------------------------  Pan ------------------------------

int tTrack::GetPan()
{
  if (Pan)
    return Pan->Value + 1;
  return 0;
}

void tTrack::SetPan(int Value)
{
  if (Pan)
    Kill(Pan);
  if (Value > 0)
  {
    tEvent *e = new tControl(0, Channel - 1, 0x0a, Value - 1);
    Put(e);
    Midi->OutNow(e);
  }
  Cleanup();
}

// ------------------------  Reverb ------------------------------

int tTrack::GetReverb()
{
  if (Reverb)
    return Reverb->Value + 1;
  return 0;
}

void tTrack::SetReverb(int Value)
{
  if (Reverb)
    Kill(Reverb);
  if (Value > 0)
  {
    tEvent *e = new tControl(0, Channel - 1, 0x5B, Value - 1);
    Put(e);
    Midi->OutNow(e);
  }
  Cleanup();
}

// ------------------------  Chorus ------------------------------

int tTrack::GetChorus()
{
  if (Chorus)
    return Chorus->Value + 1;
  return 0;
}

void tTrack::SetChorus(int Value)
{
  if (Chorus)
    Kill(Chorus);
  if (Value > 0)
  {
    tEvent *e = new tControl(0, Channel - 1, 0x5D, Value - 1);
    Put(e);
    Midi->OutNow(e);
  }
  Cleanup();
}

// ------------------------  Patch ------------------------------

int tTrack::GetPatch()
{
  if (Patch)
    return Patch->Program + 1;
  return 0;
}

void tTrack::SetPatch(int PatchNr)
{
  if (Patch)
    Kill(Patch);
  if (PatchNr > 0)
  {
    tEvent *e = new tProgram(0, Channel - 1, PatchNr - 1);
    Put(e);
    Midi->OutNow(e);
  }
  Cleanup();
}


// ------------------------  Speed ------------------------------

int tTrack::GetSpeed()
{
  if (Speed)
    return Speed->GetBPM();
  return 100;
}


void tTrack::SetSpeed(int bpm)
{
  tEvent *e = new tSetTempo(0, bpm);
  if (Speed)
    Kill(Speed);
  Put(e);
  Midi->OutNow(e);
  Cleanup();
}


// ------------------------- State ----------------------------------


char *tTrack::GetStateChar()
{
  switch (State)
  {
    case tsPlay: return "P";
    case tsMute: return "M";
    case tsSolo: return "S";
  }
  return "?";
}

void tTrack::SetState(int NewState)
{
  State = NewState % 3;
}

void tTrack::ToggleState(int Direction)
{
  State = (State + Direction + 3) % 3;
}

// ------------------------- Channel ---------------------------

void tTrack::SetChannel(int NewChannel)
{
  Channel = NewChannel;
}


